/*
Index-fu: getting the most of your indexes

Michael DeFehr - January 2013 - mike@defehr.com
*/

USE master
GO
--create table of numbers
IF EXISTS(SELECT 1 FROM sys.databases WHERE name = 'Scratch')
BEGIN
	ALTER DATABASE Scratch SET RESTRICTED_USER WITH ROLLBACK IMMEDIATE
	DROP DATABASE Scratch
END
GO
CREATE DATABASE Scratch
GO
USE Scratch
GO

IF OBJECT_ID('numbers') IS NOT NULL
BEGIN
	DROP TABLE dbo.numbers
END
ELSE
BEGIN
	CREATE TABLE numbers (
		 num int IDENTITY(1,1) NOT NULL CONSTRAINT PK_numbers PRIMARY KEY
	)
END
GO
----

DECLARE @base int, @MaxNum int
SET @MaxNum = 1000000
SET @base = 1

SET IDENTITY_INSERT dbo.numbers ON;
INSERT numbers (num) VALUES (1)

WHILE @base * 2 <= @MaxNum
BEGIN
	INSERT numbers (num)
	SELECT num + @base 
	FROM numbers
	
	SET @base = @base * 2
END

INSERT numbers (num)
SELECT num + @base 
FROM numbers
WHERE num + @base <= @MaxNum

SET IDENTITY_INSERT numbers OFF;





--Index structure
SET STATISTICS IO ON
SELECT * FROM numbers
SELECT * FROM sys.dm_db_index_physical_stats(
	DB_ID(),				-- database
	OBJECT_ID('numbers'),	-- object (table)
	1,						-- index
	NULL,					-- partition
	'DETAILED')				-- mode

SELECT * FROM numbers WHERE num = 200286

--doubly-linked list
SELECT * FROM numbers WHERE num > 800000
SELECT * FROM numbers WHERE num > 800000 ORDER BY num DESC

exec sp_helpindex 'numbers'
exec sp_sqlskills_SQL2012_helpindex 'numbers'


--composite key with text columns
USE AdventureWorks2012
GO
SELECT * FROM person.Address a

EXEC sp_SQLskills_SQL2012_helpindex 'person.address'

[AddressLine1], [AddressLine2], [City], [StateProvinceID], [PostalCode]
[AddressLine1], [AddressLine2], [City], [StateProvinceID], [PostalCode]
[AddressLine1], [AddressLine2], [City], [StateProvinceID], [PostalCode], [AddressID]

SELECT [AddressLine1], [AddressLine2], [City], [StateProvinceID], [PostalCode], AddressID FROM person.address
SELECT [AddressLine1], [AddressLine2], [City], [StateProvinceID], [PostalCode], AddressID, rowguid FROM person.address

--previous index had 622 records on a page, so we would need 19614/622 = 31 pages, and the root page would have easily held 31 records, but...

SELECT * FROM sys.dm_db_index_physical_stats(
	DB_ID(),						--database id
	OBJECT_ID('Person.Address'),	--object id
	3,								--index id
	NULL,							--partition number
	'DETAILED'						--mode (type of scan
	) ddips




--indexes are always unique
IF EXISTS(SELECT 1 FROM sys.indexes WHERE name = 'ix_Address_mddtemp') DROP INDEX ix_Address_mddtemp ON Person.Address
CREATE INDEX ix_Address_mddtemp ON person.[Address] (rowguid)

--look at sp_SQLskills_SQL2012_helpindex output
EXEC sp_SQLskills_SQL2012_helpindex 'person.address'
SELECT * FROM sys.dm_db_index_physical_stats(DB_ID(),OBJECT_ID('Person.Address'),2,NULL,'DETAILED') ddips
SELECT * FROM sys.dm_db_index_physical_stats(DB_ID(),OBJECT_ID('Person.Address'),5,NULL,'DETAILED') ddips
--space used on root page differs by just over 3% which corresponds to 256 bytes on an 8k page which is 4 bytes per tree page



--/demo




--Included Columns Demo:
USE Scratch
GO
--add some columns to the numbers table
ALTER TABLE numbers ADD Alternatekey uniqueidentifier NOT NULL DEFAULT NEWID()
ALTER TABLE numbers ADD SmallData varchar(100)
ALTER TABLE numbers ADD SomeData nvarchar(max)
GO
--put some data in the "SmallData" column
UPDATE numbers SET SmallData = CHAR(FLOOR(RAND(CHECKSUM(NEWID())) * 26) + 65)


SELECT * FROM numbers
--cannot use a LOB column in an index key
CREATE INDEX ix_numbers_SomeData ON numbers (SomeData)

--but you can get it into an index if you include it
CREATE INDEX ix_numbers_AlternateKey ON numbers(AlternateKey) INCLUDE (SomeData) 
EXEC sp_SQLskills_SQL2012_helpindex 'numbers'
DROP INDEX ix_numbers_AlternateKey ON numbers


--add columns to an index without affecting uniqueness
CREATE UNIQUE INDEX ix_numbers_AlternateKey ON numbers (AlternateKey)

SELECT TOP 100 * FROM numbers

--my unique index is there and it works - great
INSERT Numbers (Alternatekey) VALUES ('DCA0D721-22FB-41DF-81AD-FC9F653B90F9')

SELECT AlternateKey, SmallData FROM numbers WHERE AlternateKey = 'DCA0D721-22FB-41DF-81AD-FC9F653B90F9'

--what if I want SmallData in my AlternateKey index
CREATE UNIQUE INDEX ix_numbers_AlternateKey ON numbers(AlternateKey,SmallData) WITH DROP_EXISTING

BEGIN TRAN
INSERT Numbers (Alternatekey,SmallData) VALUES ('DCA0D721-22FB-41DF-81AD-FC9F653B90F9','N')
SELECT Alternatekey,SmallData FROM numbers WHERE AlternateKey = 'DCA0D721-22FB-41DF-81AD-FC9F653B90F9'
ROLLBACK

--that's not what we want - what if we include SmallData instead?
CREATE UNIQUE INDEX ix_numbers_AlternateKey ON numbers(AlternateKey) INCLUDE (SmallData) WITH DROP_EXISTING

--now AlternateKey is still unique and we don't have to do a lookup to get "SmallData"
-- (showplan on)
SELECT AlternateKey, SmallData FROM numbers WHERE AlternateKey = '534B8CC0-0A73-42D7-A302-3B0C5DF5DD57'


EXEC sp_SQLskills_SQL2012_helpindex 'numbers'
--/demo



--Filtered indexes

--how big is our index on alternate key?
SELECT * FROM sys.dm_db_index_physical_stats(DB_ID(),OBJECT_ID('numbers'),2,NULL,'DETAILED') ddips

--what if we only cared about searching by alternate key if the active flag was set?
ALTER TABLE numbers ADD ActiveFlag bit NOT NULL DEFAULT (0)
UPDATE numbers SET ActiveFlag = 1 WHERE num > 950000

SELECT * FROM numbers

CREATE UNIQUE INDEX ix_numbers_AlternateKey ON numbers(AlternateKey) INCLUDE (SmallData) WHERE ActiveFlag = 1 WITH DROP_EXISTING

--now how big is it?
SELECT * FROM sys.dm_db_index_physical_stats(DB_ID(),OBJECT_ID('numbers'),2,NULL,'DETAILED') ddips

SELECT TOP 10 AlternateKey, SmallData FROM numbers WHERE ActiveFlag = 1

SELECT AlternateKey, SmallData FROM numbers WHERE AlternateKey = '86C54563-FBC3-4F4B-A578-0003D3EB832F' AND ActiveFlag = 1


--note: uniqueness now not enforced for most of the table

INSERT numbers (AlternateKey,ActiveFlag) VALUES ('86C54563-FBC3-4F4B-A578-0003D3EB832F', 1)

BEGIN TRAN
INSERT numbers (AlternateKey,ActiveFlag) VALUES ('86C54563-FBC3-4F4B-A578-0003D3EB832F', 0)
SELECT * FROM numbers WHERE AlternateKey = '86C54563-FBC3-4F4B-A578-0003D3EB832F'
ROLLBACK


--what about unique constraints and filtered indexes?
--the gapless invoice number problem
ALTER TABLE numbers ADD InvoiceID int

CREATE UNIQUE INDEX ix_numbers_InvoiceID ON numbers(InvoiceID) WHERE InvoiceID IS NOT NULL

EXEC sp_SQLskills_SQL2012_helpindex 'numbers'
SELECT * FROM sys.dm_db_index_physical_stats(DB_ID(),OBJECT_ID('numbers'),5,NULL,'DETAILED') ddips

SELECT * FROM numbers WHERE InvoiceID = 1


IF OBJECT_ID('sq_numbers_InvoiceID') IS NOT NULL DROP SEQUENCE sq_numbers_InvoiceID
CREATE SEQUENCE sq_numbers_InvoiceID START WITH 100001 INCREMENT BY 1


IF OBJECT_ID('numbers_insert') IS NOT NULL DROP PROC numbers_insert
GO
CREATE PROC numbers_insert
@SomeData nvarchar(max),
@SmallData varchar(100),
@AlternateKey uniqueidentifier
AS
BEGIN
	BEGIN TRY
		BEGIN TRAN
			IF @AlternateKey IS NULL SET @AlternateKey = NEWID();      
			DECLARE @num int;  
			INSERT dbo.numbers
					( SomeData
					,Alternatekey
					,SmallData
					,ActiveFlag
					,InvoiceID )
			VALUES  (	@SomeData,	-- SomeData - nvarchar(max)
						@AlternateKey,	-- Alternatekey - uniqueidentifier
						@SmallData, -- SmallData - varchar(100)
						1,			-- ActiveFlag - bit
						NULL		-- InvoiceID - int
					 );
			SET @num = SCOPE_IDENTITY();
		COMMIT
	
		UPDATE dbo.numbers SET InvoiceID = NEXT VALUE FOR sq_numbers_InvoiceID WHERE num = @num;

	END TRY
	BEGIN CATCH
		IF @@TRANCOUNT > 0 ROLLBACK;
		THROW;
	END CATCH  
END
GO

EXEC dbo.numbers_insert 
    @SomeData = N'', -- nvarchar(max)
    @SmallData = 'A', -- varchar(100)
	@AlternateKey = NULL
 
SELECT * FROM dbo.numbers n WHERE n.InvoiceID IS NOT NULL  

EXEC dbo.numbers_insert 
    @SomeData = N'', -- nvarchar(max)
    @SmallData = 'A', -- varchar(100)
	@AlternateKey = 'E8A14B76-6C60-40CD-91E3-E323BF5297B8'

UPDATE numbers SET InvoiceID = 100001 WHERE InvoiceID = 100002

--/demo



-- Indexed Views

ALTER TABLE dbo.numbers ADD SumGroup int
ALTER TABLE dbo.numbers ADD Amount money
GO

UPDATE dbo.numbers SET 
	SumGroup = FLOOR(RAND(CHECKSUM(NEWID())) * 100) + 1, 
	Amount = RAND(CHECKSUM(NEWID())) * 1000

ALTER TABLE dbo.numbers ALTER COLUMN Amount money NOT NULL


SELECT TOP 100 * FROM numbers

SET STATISTICS IO ON

SELECT SumGroup, SUM(Amount)
FROM dbo.numbers n
GROUP BY SumGroup
OPTION (MAXDOP 1)
--aside: unindexed, results are unordered


--Pop quiz:
--	key columns		(SumGroup)
--	columns in tree	(SumGroup,num)
--	columns in leaf	(SumGroup,Amount,num)

IF EXISTS(SELECT 1 FROM sys.indexes WHERE name = 'ix_numbers_SumGroup') DROP INDEX ix_numbers_SumGroup ON numbers
CREATE NONCLUSTERED INDEX ix_numbers_SumGroup ON dbo.numbers (SumGroup) INCLUDE (Amount)

EXEC sp_SQLskills_SQL2012_helpindex 'numbers'

IF OBJECT_ID('qrynumbers_SumGroup') IS NOT NULL DROP VIEW qrynumbers_SumGroup
GO
CREATE VIEW qrynumbers_SumGroup
WITH SCHEMABINDING
AS
SELECT	n.SumGroup, 
		SUM(n.Amount) AS TotalAmount,
		COUNT_BIG(*) AS CountBig
FROM dbo.numbers n
GROUP BY n.SumGroup
GO

CREATE UNIQUE CLUSTERED INDEX ix_qrynumbers_SumGroup ON qrynumbers_SumGroup (SumGroup)


SELECT TOP 100 * FROM dbo.numbers n WHERE SumGroup = 10

SELECT * FROM qrynumbers_SumGroup WHERE SumGroup = 10
UPDATE numbers SET Amount = 0 WHERE num = 495
SELECT * FROM qrynumbers_SumGroup WHERE SumGroup = 10


--what about wonky constraints?

IF OBJECT_ID('PeopleThings') IS NOT NULL DROP TABLE PeopleThings
IF OBJECT_ID('People') IS NOT NULL DROP TABLE People
IF OBJECT_ID('Things') IS NOT NULL DROP TABLE Things
CREATE TABLE Things (
	ThingID int IDENTITY(1,1) CONSTRAINT PK_Things_ThingID PRIMARY KEY,
	ThingType char(10) NOT NULL,
	ThingDescription varchar(255)
);

CREATE TABLE People (
	PersonID int IDENTITY(101,1) CONSTRAINT PK_People_PersonID PRIMARY KEY,
	PersonName varchar(255)
);

CREATE TABLE PeopleThings (
	ThingID int NOT NULL CONSTRAINT FK_PeopleThings_Things FOREIGN KEY REFERENCES Things (ThingID),
	PersonID int NOT NULL CONSTRAINT FK_PeopleThings_People FOREIGN KEY REFERENCES People (PersonID),
	CONSTRAINT PK_PeopleThings PRIMARY KEY (ThingID,PersonID)
);

INSERT Things (ThingType,ThingDescription) VALUES
('PED','TV'),
('PTM','Car'),
('PD','House');

INSERT People (PersonName) VALUES
('Joe'),
('Fred'),
('Mary');

SELECT * FROM Things;
SELECT * FROM People;	

INSERT PeopleThings (ThingID, PersonID)
SELECT t.ThingID, p.PersonID
FROM Things t
	CROSS JOIN People p

-- now each person has every type of thing
SELECT * FROM Things;
SELECT * FROM People;
SELECT * FROM dbo.PeopleThings pt;

-- what if Fred has a different kind of Primary Transportation Method?
INSERT dbo.Things ( ThingType,ThingDescription )
VALUES  ('PTM','Bus'   )

SELECT * FROM Things

INSERT dbo.PeopleThings ( ThingID,PersonID )
VALUES  (4, 102)

-- so this is what Fred has, now, but he is not allowed to have more than one Primary Transportation Method!
SELECT	p.PersonName, t.ThingType, t.ThingDescription
FROM dbo.People p
	JOIN dbo.PeopleThings pt ON p.PersonID = pt.PersonID
	JOIN dbo.Things t ON pt.ThingID = t.ThingID
WHERE p.PersonID = 102

-- how to enforce this constraint?
-- obvious choice is to add ThingCode to PeopleThings and make a composite key
-- but that requires app changes and there is a design requirement to have singleton,
-- surrogate keys

-- first make DB consistent
DELETE FROM dbo.PeopleThings WHERE PersonID = 102 AND ThingID = 2

IF OBJECT_ID('qryPeopleThings_Things') IS NOT NULL DROP VIEW qryPeopleThings_Things
GO
CREATE VIEW qryPeopleThings_Things
WITH SCHEMABINDING
AS
SELECT	pt.PersonID,
		t.ThingType
FROM dbo.PeopleThings pt
	JOIN dbo.Things t ON pt.ThingID = t.ThingID
GO

CREATE UNIQUE CLUSTERED INDEX ix_qryPeopleThings_Things ON qryPeopleThings_Things (PersonID,ThingType)

INSERT dbo.PeopleThings ( ThingID,PersonID )
VALUES  (2, 102)

DELETE FROM PeopleThings WHERE PersonID = 102 AND ThingID = 4 

INSERT dbo.PeopleThings ( ThingID,PersonID )
VALUES  (2, 102)

INSERT dbo.Things ( ThingType,ThingDescription )
VALUES  ('PCD','Microwave'   )

INSERT dbo.PeopleThings ( ThingID,PersonID )
VALUES  (5, 102)